## a modified version of all.models that stops searching for models of
## higher numbers of steps once they are no longer significant given a
## p value of 0.05 and significance level of sig.level (using a chi
## square test with 1 degree of freedom to test for significance)
fit.mosaic <- function(data, sig.level, N=50, ngens=100, s=5,
                       p.mutate=0.1, p.sex=1, p.rec=0.25,
                       verbose=TRUE) {
  ## Run a model with k steps until we return a likelihood that is
  ## better than 'target'
  next.model <- function(k) {
    MAXREPS <- 3
    for ( i in 1:MAXREPS ) {
      ans <- mosaic(data, logLik, k, N, ngens, s, p.mutate, p.sex, p.rec,
                    verbose=verbose)
      if ( (ans$best.ll - res[[k-1]]$best.ll) > -0.001 )
        return(ans)
      else if ( verbose > 1 )
        cat("oops! failed to converge...\n")
    }
    cat("Failed to find a better model; probably rounding error\n")
    ans
  }

  if ( verbose )
    cat("Preparing likelihood function\n")
  logLik <- make.mosaic.logLik(data)
  
  chiSquareValue <- qchisq(sig.level, df=2, lower.tail=FALSE)/2
  n.patches <- nrow(data)
  n.partitions <- seq_len(n.patches-1)
  
  res <- vector("list", n.patches-1)
  for ( i in n.partitions ) {

    if ( i == 1 )
      res[[1]] <- mosaic(data, logLik, i,  N, ngens, s, p.mutate, p.sex,
                         p.rec, verbose=verbose)
    else {
      res[[i]] <- next.model(i)
      is.signif <- res[[i]]$best.ll-res[[i-1]]$best.ll > chiSquareValue
      if ( !is.signif ) {
        if ( verbose > 1 )
          cat("Second attempt; have we topped out?\n")
        res2 <- next.model(i)
        if ( res2$best.ll - res[[i-1]]$best.ll > chiSquareValue )
          res[[i]] <- res2
        else {
          if ( verbose > 1 )
            cat(sprintf("\nFitted %d/%d steps.\n", i, n.patches))
          break
        }
      }
    }
  }
  ret <- res[1:(i-1)]
  attr(ret, "data") <- data
  class(ret) <- "mosaic"
  ret
}

## Computes mosaic statistic for a given number of partitions.
mosaic <- function(data, logLik, n.partitions,
                   N, ngens, s, p.mutate, p.sex, p.rec,
                   max.exhaust=N*ngens*2, verbose=TRUE) {
  n.patches <- nrow(data)
  n.exhaust <- choose(n.patches-1, n.partitions)

  if ( n.exhaust > max.exhaust ) {
    if ( verbose )
      cat(sprintf("GA: (%d/%d)\n", n.partitions, n.patches))
    mosaic.ga(data, ngens, n.partitions, N, s, p.mutate, p.sex, p.rec,
              logLik)
  } else {
    if ( verbose )
      cat(sprintf("Exhaust: (%d/%d)\n", n.partitions, n.patches))
    mosaic.exhaust(data, n.partitions, logLik)
  }
}

## Exhaustively search for n.partitions
mosaic.exhaust <- function(data, n.partitions, logLik) {
  n.patches <- nrow(data)
  at <- combn(n.patches-1, n.partitions, simplify=FALSE)
  ll <- sapply(at, logLik)
  i <- which.max(ll)
  list(best.ll=ll[i], best.model=at[[i]], best.ll.t=ll[i])
}
