#### (1) Initial state
## initial.popn randomly generates N solutions for n.patches patches
## divided into n.partitions partitions.  The function logLik function
## computes the log likelihood given a vector of partitions.
initial.popn <- function(N, n.partitions, n.patches, logLik) {
  popn <- lapply(seq_len(N), function(x)
                 sort(sample(n.patches-1, n.partitions)))
  w <- sapply(popn, logLik)
  list(popn=popn,
       w=w,
       best.w=max(w),
       best.model=popn[[which.max(w)]])  
}

#### (2) Mutation
## Next, take a population 'popn', with a fitness vector 'w' (where
## the ith element in 'w' gives the fitness of the ith model in
## 'popn').
##
## First, draw binomial random deviates to determine how many
## mutations will happen per model (mutations happen at an individual
## partition with probability 'p.mutate', and there are
## 'n.partitions'=length(popn[[1]]) partitions available for
## mutation).
##
## Mutate the models that had at least one mutation, by passing them
## through to 'mutate'.
##
## Only mutated models need their fitness recalculated.
##
## Finally, return a list with the new population of models and their
## fitness vector.
mutate <- function(popn, w, n.patches, p.mutate, logLik) {
  nmutants <- rbinom(length(popn), length(popn[[1]]), p.mutate)

  mutated <- which(nmutants>0)

  if ( length(mutated) > 0 ) {
    popn[mutated] <-
      mapply(mutate1, popn[mutated], nmutants[mutated], n.patches,
             SIMPLIFY=FALSE)
    w[mutated] <- sapply(popn[mutated], logLik)
  }
  list(popn=popn, w=w)
}

## Take a model 'x', describing a hybrid zone of up to 'n.patches'
## patches, mutate 'n.mutants' of the steps.  To do this, draw from
## all possible step locations (on 1:(n.patches-1)) without the steps
## that are in use by non mutated steps (i.e. a step can mutate to
## itself).
mutate1 <- function(x, n.mutants, n.patches) {
  mutated <- sort(sample(length(x), n.mutants))
  x[mutated] <-
    sample(setdiff(seq_len(n.patches-1), x[-mutated]), n.mutants)
  sort(x)
}

#### (3) Selection:
## This selection regime is fairly straightforward; sample with
## replacement s*N models from the population, then take the N best
## models.
select <- function(popn, w, s) {
  N <- length(popn)
  i <- sample(N, ceiling(s*N), TRUE)
  take <- i[order(w[i], decreasing=TRUE)[1:N]]
  list(popn=popn[take], w=w[take])
}

select.weighted <- function(popn, w) {
  N <- length(popn)
  take <- sample(N, N, TRUE, w-min(w)+1)
  list(popn=popn[take], w=w[take])
}

#### (4) Recombiation
## Select a fraction of the population to be recombined.  These are
## recombined with randomly selected individuals in the previous
## generation (popn0).  The fitness of the recombinants is then
## calculated.
recombine <- function(state1, popn0, p.sex, p.rec, n.patches,
                      logLik) {
  popn1 <- state1$popn
  w1 <- state1$w
  N <- length(popn1)
  nrec <- rbinom(1, N, p.sex)
  i <- sample(N, nrec)
  if ( length(i) > 0 ) {
    b <- popn0[sample(N, nrec, TRUE)]
    popn1[i] <- mapply(recombine1, popn1[i], b, p=p.rec,
                       n.patches=n.patches, SIMPLIFY=FALSE)
    w1[i] <- sapply(popn1[i], logLik)
  }
  list(popn=popn1, w=w1)
}

## Recombination function:
recombine1 <- function(a, b, p, n.patches) {
  a0 <- a
  rec <- runif(length(a)) < p
  if ( any(rec) ) {
    avail <- setdiff(b, a[!rec])

    if ( length(avail) == 1 )
      a[rec] <- avail
    else 
      a[rec] <- sample(avail, sum(rec))

    sort(a)
  } else {
    a
  }
}

#### (5) Step the population through one full generation:
step <- function(state, s, n.patches, p.mutate, p.sex, p.rec,
                 logLik) {
  ## Mutation
  state1 <-
    update.state(state,
                 mutate(state$popn, state$w, n.patches, p.mutate,
                        logLik))
  ## Selection
  state2 <- select(state1$popn, state1$w, s)
  ## Recombination
  update.state(state1,
               recombine(state2, state1$popn, p.sex, p.rec, n.patches,
                         logLik))
} 

## Bookkeeping to make sure that we keep the best model
update.state <- function(state0, state1) {
  if ( max(state1$w) > state0$best.w ) {
    state1$best.w     <- max(state1$w)
    state1$best.model <- state1$popn[[which.max(state1$w)]]
  } else {
    state1$best.w     <- state0$best.w
    state1$best.model <- state0$best.model
  }
  state1
}

## Run the GA for ngens generations, collecting the best fitness over
## time, and returning the state and best model.
run <- function(x, ngens, s, n.patches, p.mutate, p.sex, p.rec,
                logLik) {
  out <- vector("numeric", ngens)
  for ( i in seq_len(ngens) ){
    x <- step(x, s, n.patches, p.mutate, p.sex, p.rec, logLik)
    out[i] <- x$best.w
  }
  list(best.ll=x$best.w, best.model=x$best.model, best.ll.t=out)
}

## Wrapper around run(), that generates the initial population and the
## logLik function.
mosaic.ga <- function(data, ngens, n.partitions, N, s, p.mutate,
                      p.sex, p.rec, logLik) {
  n.patches <- nrow(data)
  run(initial.popn(N, n.partitions, n.patches, logLik),
      ngens, s, n.patches, p.mutate, p.sex, p.rec, logLik)
}
